use crate::{config::Context, ip::BlockIpData, ipblc::PKG_NAME};

use std::{
    io::Error,
    net::{IpAddr, Ipv4Addr, Ipv6Addr},
    sync::Arc,
};

use rustables::{expr::*, *};
use tokio::sync::RwLock;

pub enum FwTableType {
    IPv4,
    IPv6,
}

#[allow(dead_code)]
pub enum FwAction {
    Add,
    Delete,
}

macro_rules! initrules {
    ($batch:expr, $table:expr, $chain:ident, $reset:expr) => {
        $chain.set_hook(Hook::new(HookClass::In, 1));

        $batch.add(&$chain, MsgType::Add);
        if $reset {
            $batch.add(&Rule::new(&$chain).unwrap(), MsgType::Del);
        }
    };
}

macro_rules! makerules {
    ($ipdata:ident, $chain:ident, $batch:ident, $t:ty, $ip_t:ident,$action:ty) => {
        let ip = $ipdata.ipdata.ip.parse::<$t>().unwrap();
        Rule::new(&$chain)
            .unwrap()
            .saddr(ip.into())
            .with_expr(Conntrack::new(ConntrackKey::State))
            .with_expr(Bitwise::new(10u32.to_le_bytes(), 0u32.to_be_bytes()).unwrap())
            .with_expr(Cmp::new(CmpOp::Neq, 0u32.to_be_bytes()))
            .with_expr(Counter::default())
            .drop()
            .add_to_batch(&mut $batch);
    };
}

pub fn fwglobalinit(t: FwTableType, reset: bool) -> (Batch, Chain) {
    let table_name: String;
    let table: Table;
    let mut chain: Chain;
    match t {
        FwTableType::IPv4 => {
            table_name = format!("{PKG_NAME}4");
            table = Table::new(ProtocolFamily::Inet).with_name(table_name);
            chain = Chain::new(&table)
                .with_policy(ChainPolicy::Accept)
                .with_name(PKG_NAME);
        }
        FwTableType::IPv6 => {
            table_name = format!("{PKG_NAME}6");
            table = Table::new(ProtocolFamily::Inet).with_name(table_name);
            chain = Chain::new(&table)
                .with_policy(ChainPolicy::Accept)
                .with_name(PKG_NAME);
        }
    }
    let mut batch = Batch::new();

    batch.add(&table, MsgType::Add);
    initrules!(batch, table, chain, reset);

    (batch, chain)
}

pub fn fwblock<'a>(ip_add: &BlockIpData) -> std::result::Result<&String, error::QueryError> {
    let (mut batch4, chain4) = fwglobalinit(FwTableType::IPv4, false);
    let (mut batch6, chain6) = fwglobalinit(FwTableType::IPv6, false);

    match ip_add.ipdata.t {
        4 => {
            makerules!(ip_add, chain4, batch4, Ipv4Addr, ipv4, FwAction::Add);
            match batch4.send() {
                Ok(_) => {}
                Err(e) => {
                    println!("block not ok {e} {ip_add:?}")
                }
            }
        }
        6 => {
            makerules!(ip_add, chain6, batch6, Ipv6Addr, ipv6, FwAction::Add);
            match batch6.send() {
                Ok(_) => {}
                Err(e) => {
                    println!("block not ok {e} {ip_add:?}")
                }
            }
        }
        _ => {}
    }

    Ok(&ip_add.ipdata.ip)
}

pub fn fwunblock<'a>(ip_del: &BlockIpData) -> std::result::Result<&String, error::QueryError> {
    let (mut batch4, chain4) = fwglobalinit(FwTableType::IPv4, false);
    let (mut batch6, chain6) = fwglobalinit(FwTableType::IPv6, false);

    match ip_del.ipdata.t {
        4 => {
            for h in &ip_del.handle {
                let r = Rule::new(&chain4).unwrap().with_handle(*h);
                batch4.add(&r, MsgType::Del);
            }
            match batch4.send() {
                Ok(_) => {}
                Err(e) => {
                    println!("delete not ok {e} {ip_del:?}")
                }
            }
        }
        6 => {
            for h in &ip_del.handle {
                let r = Rule::new(&chain6).unwrap().with_handle(*h);
                batch6.add(&r, MsgType::Del);
            }
            match batch6.send() {
                Ok(_) => {}
                Err(e) => {
                    println!("delete not ok {e} {ip_del:?}")
                }
            }
        }
        _ => {}
    }
    Ok(&ip_del.ipdata.ip)
}

pub async fn get_current_rules(
    ctx: &Arc<RwLock<Context>>,
    ret: &mut Vec<String>,
    fwlen: &mut usize,
) -> Result<(), Error> {
    let mut ips_all_count = 0;
    let tables: &[String] = &[format!("{PKG_NAME}4"), format!("{PKG_NAME}6")];
    for table_name in tables {
        let get_table = || -> Result<Option<Table>, Error> {
            let tables = list_tables().unwrap();
            for table in tables {
                if let Some(name) = table.get_name() {
                    if name == table_name {
                        return Ok(Some(table));
                    }
                }
            }

            Ok(None)
        };

        let get_chain = |table: &Table| -> Result<Option<Chain>, Error> {
            let chains = list_chains_for_table(table).unwrap();
            for chain in chains {
                if let Some(name) = chain.get_name() {
                    if *name == "ipblc" {
                        return Ok(Some(chain));
                    }
                }
            }

            Ok(None)
        };

        let table = get_table()?.expect("no table?");
        let chain = get_chain(&table)?.expect("no chain?");

        let mut ctx = { ctx.write().await };
        let rules = list_rules_for_chain(&chain).unwrap().clone();

        for (ip, b) in ctx.blocklist.iter_mut() {
            let ip_parsed: IpAddr = ip.parse().unwrap();

            let cmprule = Rule::new(&chain).unwrap().saddr(ip_parsed).drop();
            let mut gexpr = RawExpression::default();
            for e in cmprule.get_expressions().unwrap().iter() {
                if let Some(ExpressionVariant::Cmp(_ip)) = e.get_data() {
                    gexpr = e.clone();
                }
            }

            for rule in rules.iter() {
                for expr in rule.get_expressions().unwrap().iter() {
                    if let Some(expr::ExpressionVariant::Cmp(_)) = expr.get_data() {
                        if gexpr == expr.clone() {
                            b.handle.insert(*rule.get_handle().unwrap());
                            ips_all_count += 1;
                        }
                    }
                }
            }
        }
    }

    if *fwlen != ips_all_count {
        ret.push(format!("{length} ip in firewall", length = ips_all_count));
    }
    *fwlen = ips_all_count;

    Ok(())
}

#[allow(dead_code)]
fn fw_rules_count() -> i64 {
    0
}
